"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
require("tocca");
const ko = require("knockout");
const style_1 = require("./style");
const eventListener_1 = require("./eventListener");
const userAgent_1 = require("./userAgent");
const uiElementPropertyName = "uiElement";
const mouseEvents = {
    contextmenu: true,
    mousedown: true,
    mouseenter: true,
    mouseleave: true,
    mousemove: true,
    mouseout: true,
    mouseover: true,
    mouseup: true
};
const touchEvents = {
    touchcancel: true,
    touchend: true,
    touchmove: true,
    touchstart: true,
    longtap: true,
    swipeleft: true,
    swiperight: true,
    swipeup: true,
    swipedown: true
};
function isGlobalEvent(event) {
    return !(mouseEvents[event] || touchEvents[event]);
}
exports.settings = {
    tapDelay: 300
};
let observerOptions = {
    childList: true,
    subtree: true
};
const mutationObserver = new MutationObserver((mutationRecords) => {
    mutationRecords.forEach((mutationRecord) => {
        if (mutationRecord.removedNodes) {
            mutationRecord.removedNodes.forEach((removedNode) => {
                const uiElement = removedNode[uiElementPropertyName];
                if (uiElement && uiElement.disposeOnDomRemoval && uiElement.dispose) {
                    uiElement.dispose();
                }
            });
        }
    });
});
mutationObserver.observe(document.body, observerOptions);
function setObserverOptions(options) {
    observerOptions = options;
}
exports.setObserverOptions = setObserverOptions;
function isNotNullOrUndefined(value) {
    return !(value === null || typeof value === "undefined");
}
class UiElement {
    constructor(tagName, elementType, options) {
        this._attributes = {};
        this._classes = {};
        this._style = {};
        this._userSelectable = true;
        this._addControlToDataDictionary = false;
        this._children = new Array();
        this._knockoutSubscriptions = Array();
        this._styleDisplayVisible = "";
        this._visible = true;
        this.diposeOnDomRemoval = false;
        this.dispose = () => {
            this.remove();
            if (this._children) {
                this._children.forEach((child) => {
                    child.dispose();
                });
            }
            this._knockoutSubscriptions.forEach((subscription) => {
                subscription.dispose();
            });
        };
        this._tagName = tagName;
        this._elementType = elementType || "UiElement";
        this.options = options || {};
        if (options) {
            if (isNotNullOrUndefined(options.id)) {
                this.setAttribute("id", options.id);
            }
            if (options.elementType) {
                this._elementType = ko.unwrap(options.elementType);
                if (ko.isObservable(options.elementType)) {
                    this.addSubscription(options.elementType, () => {
                        this.setElementType(elementType);
                    });
                }
            }
            if (options.classNames) {
                this.setClasses(ko.unwrap(options.classNames), true);
                if (ko.isObservable(options.classNames)) {
                    this.addSubscription(options.classNames, (classNames) => {
                        this.setClasses(classNames, true);
                    });
                }
            }
            if (options.attributes) {
                this._setAttributes(ko.unwrap(options.attributes));
                if (ko.isObservable(options.attributes)) {
                    this.addSubscription(options.attributes, (attributes) => {
                        this._setAttributes(attributes);
                    });
                }
            }
            if (options.eventListeners) {
                this._eventListeners = options.eventListeners;
            }
            if (options.style) {
                for (const styleProperty of Object.keys(options.style)) {
                    const styleValue = options.style[styleProperty];
                    this._style[styleProperty] = ko.unwrap(styleValue);
                    if (ko.isObservable(styleValue)) {
                        this.addSubscription(styleValue, (newValue) => {
                            this.setStyle({ [styleProperty]: newValue });
                        });
                    }
                }
            }
            if (options.userSelectable === false) {
                this._userSelectable = false;
            }
            if (isNotNullOrUndefined(options.innerHtml)) {
                this.setInnerHtml(ko.unwrap(options.innerHtml));
                if (ko.isObservable(options.innerHtml)) {
                    this.addSubscription(options.innerHtml, (innerHtml) => {
                        this.setInnerHtml(innerHtml);
                    });
                }
            }
            if (isNotNullOrUndefined(options.innerText)) {
                this.setInnerText(ko.unwrap(options.innerText));
                if (ko.isObservable(options.innerText)) {
                    this.addSubscription(options.innerText, (newValue) => {
                        this.setInnerText(newValue);
                    });
                }
            }
            if (options.addControlToDataDictionary) {
                this._addControlToDataDictionary = true;
            }
            if (options.children) {
                this._children = ko.unwrap(options.children);
                if (ko.isObservable(options.children)) {
                    this.addSubscription(options.children, (newChildren) => {
                        this.setChildren(newChildren);
                    });
                }
            }
            if (options.dispose) {
                var dispose = this.dispose;
                this.dispose = () => {
                    dispose();
                    options.dispose();
                };
            }
            if (options.refresh) {
                this.refresh = options.refresh;
            }
            var visible = ko.unwrap(options.visible);
            if (visible === false) {
                this._visible = false;
            }
            if (ko.isObservable(options.visible)) {
                this.addSubscription(options.visible, (newValue) => {
                    this.setVisibility(newValue);
                });
            }
        }
    }
    _addEventListenerToElement(type, listener, options) {
        if (!type) {
            throw new Error("type is mandatory");
        }
        if (!listener) {
            throw new Error("listener is mandatory");
        }
        if (!options) {
            options = {};
        }
        if (!options.passive) {
            options.passive = false;
        }
        if (!options.once) {
            options.once = false;
        }
        if (this.element.addEventListener) {
            this.element.addEventListener(type, (event) => {
                listener(event);
            }, {
                passive: options.passive,
                once: options.once
            });
        }
        else if (this.element.attachEvent) {
            this.element.attachEvent(type, (event) => {
                listener(event);
            });
        }
        else {
            throw "Your browser does not support 'addEventListener'";
        }
    }
    _setAttributes(attributes) {
        const existingAttributeNames = Object.keys(this._attributes);
        this._attributes = {};
        attributes.forEach((attribute) => {
            this._attributes[attribute.name] = ko.unwrap(attribute.value);
            if (ko.isObservable(attribute.value)) {
                this.addSubscription(attribute.value, (newValue) => {
                    this.setAttribute(attribute.name, newValue);
                });
            }
        });
        if (existingAttributeNames) {
            existingAttributeNames.forEach((existingAttributeName) => {
                if (!this._attributes[existingAttributeName]) {
                    this.removeAttribute(existingAttributeName);
                }
            });
        }
    }
    build() {
        if (this._elementType) {
            this.setAttribute("elementtype", this._elementType);
        }
        const display = this.getStyleValue("display");
        if (display && display !== "none") {
            this._styleDisplayVisible = display;
        }
        this.setVisibility(this._visible);
        ko.applyBindingsToNode(this.element, {
            style: this._style,
            attr: this._attributes
        });
        if (isNotNullOrUndefined(this._innerHtml)) {
            this.element.innerHTML = this._innerHtml;
        }
        if (isNotNullOrUndefined(this._innerText)) {
            this.element.innerText = this._innerText;
        }
        this.element.className = Object.keys(this._classes).join(" ");
        if (this._userSelectable === false) {
            this._addEventListenerToElement("selectstart", () => { return false; }, { passive: true });
            this.element.style.userSelect = "none";
        }
        if (this._children) {
            this._children.forEach((child) => {
                this.element.appendChild(child.render());
            });
        }
        if (this._addControlToDataDictionary) {
            if (!this.element.data) {
                this.element.data = {};
            }
            this.element.data.uiElement = this;
        }
    }
    render() {
        if (this.element) {
            this.remove();
        }
        if (typeof this.build === "function") {
            this.element = document.createElement(this._tagName);
            this.build();
            const clickEventListeners = new Array();
            const doubleClickEventListeners = new Array();
            if (this._eventListeners) {
                const eventListeners = this._eventListeners;
                eventListeners.forEach((eventListener) => {
                    if (eventListener.type === "click") {
                        clickEventListeners.push(eventListener);
                        return;
                    }
                    if (eventListener.type === "dblclick") {
                        doubleClickEventListeners.push(eventListener);
                        return;
                    }
                    if (eventListener.type === "dragstart") {
                        this.setAttribute("draggable", "true");
                    }
                    if (isGlobalEvent(eventListener.type)) {
                        this._addEventListenerToElement(eventListener.type, (event) => {
                            if (eventListener.options) {
                                const options = eventListener.options;
                                const keyEvent = event;
                                if (options.altKey && !keyEvent.altKey) {
                                    return;
                                }
                                if (options.shiftKey && !keyEvent.shiftKey) {
                                    return;
                                }
                                if (options.ctrlKey && !keyEvent.ctrlKey) {
                                    return;
                                }
                                if (options.eventKey && options.eventKey !== keyEvent.key) {
                                    return;
                                }
                            }
                            eventListener.listener.call(this, event);
                        }, eventListener.options);
                    }
                    else if (mouseEvents[eventListener.type] && userAgent_1.userAgent.device.supportsMouseEvents) {
                        this._addEventListenerToElement(eventListener.type, eventListener.listener, eventListener.options);
                    }
                    else if (touchEvents[eventListener.type] && userAgent_1.userAgent.device.supportsTouchEvents) {
                        this._addEventListenerToElement(eventListener.type, eventListener.listener, eventListener.options);
                    }
                });
                clickEventListeners.forEach((clickEventListener) => {
                    if (userAgent_1.userAgent.device.supportsTouchEvents) {
                        this._addEventListenerToElement("tap", clickEventListener.listener, clickEventListener.options);
                    }
                    if (userAgent_1.userAgent.device.supportsMouseEvents) {
                        this._addEventListenerToElement(clickEventListener.type, clickEventListener.listener, clickEventListener.options);
                    }
                });
                doubleClickEventListeners.forEach((doubleClickEventListener) => {
                    if (userAgent_1.userAgent.device.supportsTouchEvents) {
                        this._addEventListenerToElement("dbltap", doubleClickEventListener.listener, doubleClickEventListener.options);
                    }
                    if (userAgent_1.userAgent.device.supportsMouseEvents) {
                        this._addEventListenerToElement(doubleClickEventListener.type, doubleClickEventListener.listener, doubleClickEventListener.options);
                    }
                });
            }
            return this.element;
        }
        throw "The build method of this UiElement has not been defined";
    }
    refresh() { }
    remove() {
        if (this.element) {
            if (this.element.remove) {
                this.element.remove();
            }
            else {
                if (this.element.parentElement) {
                    this.element.parentElement.removeChild(this.element);
                }
            }
            this.element = null;
        }
    }
    handleMessage(message) {
        console.log(JSON.stringify(message));
    }
    getElement() {
        return this.element;
    }
    focus(options) {
        if (this.element) {
            this.element.focus(options);
        }
    }
    focusPreviousSibling(options) {
        if (this.element && this.element.previousElementSibling && this.element.nextSibling instanceof HTMLElement) {
            this.element.nextSibling.focus(options);
        }
    }
    focusNextSibling(options) {
        if (this.element && this.element.previousElementSibling && this.element.nextSibling instanceof HTMLElement) {
            this.element.nextSibling.focus(options);
        }
    }
    setElementType(elementType) {
        this._elementType = elementType;
        this.setAttribute("elementType", elementType);
    }
    setClasses(classNames, replace) {
        if (replace) {
            this._classes = {};
        }
        if (classNames) {
            classNames.forEach((className) => {
                this._classes[className] = true;
            });
        }
        if (this.element) {
            this.element.className = Object.keys(this._classes).join(" ");
        }
    }
    removeClasses(classNames) {
        if (classNames) {
            classNames.forEach((className) => {
                if (this._classes[className]) {
                    delete this._classes[className];
                }
                if (this.element) {
                    this.element.classList.remove(className);
                }
            });
        }
    }
    setInnerHtml(innerHtml) {
        if (isNotNullOrUndefined(innerHtml)) {
            this.setInnerText(null);
        }
        this._innerHtml = innerHtml;
        if (this.element) {
            this.element.innerHTML = innerHtml;
        }
    }
    setInnerText(innerText) {
        if (isNotNullOrUndefined(innerText)) {
            this.setInnerHtml(null);
        }
        this._innerText = innerText;
        if (this.element) {
            this.element.innerText = innerText;
        }
    }
    setAttribute(name, value) {
        this._attributes[name] = value;
        if (this.element) {
            this.element.setAttribute(name, value);
        }
    }
    removeAttribute(attributeName) {
        if (this._attributes[attributeName]) {
            delete this._attributes[attributeName];
        }
        if (this.element) {
            this.element.removeAttribute(attributeName);
        }
    }
    getAttribute(attributeName) {
        if (this.element) {
            return this.element.getAttribute(attributeName);
        }
        else {
            return this._attributes[attributeName];
        }
    }
    getStyle(...cssProperties) {
        if (!cssProperties || cssProperties.length === 0) {
            return this._style;
        }
        const style = {};
        if (cssProperties) {
            cssProperties.forEach((cssProperty) => {
                if (this._style[cssProperty]) {
                    style[cssProperty] = this._style[cssProperty];
                }
                else if (this.element) {
                    style[cssProperty] = this.element.style[cssProperty];
                }
            });
        }
        return style;
    }
    getStyleValue(cssProperty) {
        if (this._style[cssProperty]) {
            return this._style[cssProperty];
        }
        else if (this.element) {
            return this.element.style[cssProperty];
        }
        return null;
    }
    setStyle(newStyle, overwriteExisting) {
        if (newStyle) {
            Object.keys(newStyle).forEach((key) => {
                if (this._style.hasOwnProperty(key)) {
                    if (!overwriteExisting) {
                        return;
                    }
                }
                const newValue = newStyle[key];
                this._style[key] = newValue;
                if (this.element) {
                    if (newValue === null || newValue === undefined) {
                        if (userAgent_1.userAgent.browser.isInternetExplorer) {
                            this.element.style[key] = style_1.defaultStyle[key];
                        }
                        else {
                            this.element.style[key] = null;
                        }
                    }
                    else {
                        this.element.style[key] = newValue;
                    }
                }
            });
        }
    }
    setStyleNonStandard(name, value, overwriteExisting) {
        this.setStyle({ [name]: value }, overwriteExisting);
    }
    deleteStyle(style) {
        let styles;
        if (style instanceof Array) {
            styles = style;
        }
        else if (style instanceof Object) {
            styles = Object.keys(style);
        }
        else {
            styles = [style];
        }
        styles.forEach((cssProperty) => {
            if (this._style && this._style.hasOwnProperty(cssProperty)) {
                delete this._style[cssProperty];
            }
            if (this.element) {
                this.element.style[cssProperty] = undefined;
            }
        });
    }
    setVisibility(visible) {
        this._visible = visible;
        if (this.element) {
            if (visible) {
                this.element.style.display = this._styleDisplayVisible;
            }
            else {
                this.element.style.display = "none";
            }
        }
    }
    addSubscription(observable, callback) {
        this._knockoutSubscriptions.push(observable.subscribe(callback));
    }
    dispatchEvent(type) {
        if (this.element) {
            return this.element.dispatchEvent(new Event(type));
        }
        return false;
    }
    addEventListener(type, listener, options) {
        if (!this._eventListeners) {
            this._eventListeners = new Array();
        }
        this._eventListeners.push(new eventListener_1.EventListener(type, listener, options));
    }
    getChildren() {
        return this._children;
    }
    setChildren(children, replace = true) {
        if (!children) {
            children = [];
        }
        if (replace) {
            this._children = children;
            if (this.element) {
                while (this.element.firstChild) {
                    this.element.removeChild(this.element.firstChild);
                }
            }
        }
        else {
            this._children.push(...children);
        }
        if (this.element) {
            if (children.length > 0) {
                const documentFragment = document.createDocumentFragment();
                children.forEach((child) => {
                    documentFragment.appendChild(child.render());
                });
                this.element.appendChild(documentFragment);
            }
        }
    }
    addChild(newChild, referenceChild) {
        if (referenceChild) {
            const index = this._children.indexOf(referenceChild);
            if (index > -1) {
                this._children.splice(index, 0, newChild);
                if (this.element && referenceChild.getElement()) {
                    this.element.insertBefore(newChild.render(), referenceChild.getElement());
                }
            }
        }
        else {
            this._children.push(newChild);
            if (this.element) {
                this.element.appendChild(newChild.render());
            }
        }
    }
    removeChild(element) {
        const index = this._children.indexOf(element);
        if (index > -1) {
            this._children.splice(index, 1);
            if (this.element) {
                this.element.removeChild(this.element.children[index]);
            }
        }
    }
    empty() {
        this._innerText = null;
        this._innerHtml = null;
        this._children = [];
        if (this.element) {
            this.element.innerHTML = "";
            while (this.element.firstChild) {
                this.element.removeChild(this.element.firstChild);
            }
        }
    }
}
exports.UiElement = UiElement;
//# sourceMappingURL=uiElement.js.map